#include "eprom.h"

/*
 * Ermittelt den EPROM-Typ anhand der Schalterstellung auf der
 * Platine.
 */
uint8_t eprom_get_type(void) {

    // Falls eine 1 an Pin EPEEP anliegt (der Schalter steht
    // auf EPROM), ist das EPROM ein (normales) EPROM.
    if (PIND & (1<<EPEEP)) {
        return EPROM_NEPROM;
    } else { 
        return EPROM_EEPROM;
    }

}

/*
 * Liest den Wert, der an der uebergebenen Adresse im EPROM
 * gespeichert ist, aus und gibt ihn zurueck.
 */
uint8_t eprom_read(uint16_t address) {

    // Das EPROM soll auf den LCD-/EPROM-Bus schreiben.
    misc_set_data_direction(MISC_DD_EPROM);

    // Die Adresse wird auf die Adressleitungen gelegt.
    PORTB = address & 0xff;
    PORTD &= ~(0x70);
    PORTD |= (address>>4) & 0x70;

    // Stabilisierung abwarten.
    _delay_us(10);

    // Wert lesen und zurueckgeben.
    return PINA;
}

/*
 * Speichert, ob der Brennvorgang initialisiert wurde und
 * welcher EPROM-Typ gebrannt wird.
 */
uint8_t eprom_type = EPROM_NOEPROM;

/*
 * Initialisiert den Brennvorgang, muss vor dem ersten Aufruf
 * von eprom_write() aufgerufen werden.
 */
void eprom_write_start(void) {

    // Falls eprom_write_start() bereits aufgerufen wurde 
    // (ohne zwischenzeitlich eprom_write_stop() aufzurufen),
    // erzeuge eine Fehlermeldung.
    if (eprom_type != EPROM_NOEPROM) {
        misc_error();
        return;
    }

    if (eprom_get_type() == EPROM_NEPROM) { 

        // Setze den Typ des EPROMs und aktiviere die
        // Brennspannung fuer das EPROM via VPPWE.
        eprom_type = EPROM_NEPROM;
        PORTD &= ~(1<<VPPWE);

    } else { 
        
        // Setze PROM-Typ auf EPROM, keine Beeinflussung der
        // Brennspannung noetig
        eprom_type = EPROM_EEPROM;

    }

    // Warte, bis die Brennspannung korrekt anliegt.
    _delay_ms(250);

}

/*
 * Beendet den Brennvorgang, sollte nach dem letzten Aufruf von
 * eprom_write() aufgerufen werden.
 */
void eprom_write_stop(void) {

    switch (eprom_type) {

        case EPROM_NEPROM: // EPROM

            // Deaktiviere Brennspannung und setze den Typ des
            // PROMs zurueck.
            PORTD |= (1<<VPPWE);
            eprom_type = EPROM_NOEPROM;
            break;

        case EPROM_EEPROM: // EEPROM

            // Setze den Typ des PROMs zurueck.
            eprom_type = EPROM_NOEPROM;
            break;

        // eprom_write_stop() wurde aufgerufen, ohne dass
        // vorher eprom_write_start() aufgerufen wurde.
        default:
            misc_error();
            break;

    }

}

/*
 * Brennt den uebergebenen Wert an die uebergebene Adresse,
 * eprom_write_start() muss vor dem ersten Aufruf aufgerufen
 * werden.
 */
void eprom_write(uint16_t address, uint8_t data) {

    // Der Controller soll auf den LCD-/EPROM-Bus schreiben.
    misc_set_data_direction(MISC_DD_UC);

    // Die Adresse wird auf die Adress- und der Wert auf die
    // Datenleitungen gelegt.
    PORTB = address & 0xff;
    PORTD &= ~(0x70);
    PORTD |= (address>>4) & 0x70;
    PORTA = data;

    // Stabilisierung abwarten.
    _delay_us(10);

    switch (eprom_type) {

        case EPROM_NEPROM: //EPROM

            // Brennimpuls fuer 50ms anlegen.
            PORTC |= (1<<EPCE);
            _delay_ms(50);
            PORTC &= ~(1<<EPCE);
            break;

        case EPROM_EEPROM: // EEPROM

            // Brennimpuls geben und Stabilisierung abwarten.
            PORTD &= ~(1<<VPPWE);
            _delay_us(10);
            PORTD |= (1<<VPPWE);
            break;

        // eprom_write() wurde aufgerufen, ohne dass
        // vorher eprom_write_start() aufgerufen wurde.
        default:
            misc_error();
            break;

    }

}

/*
 * Liest das EPROM aus und gibt zurueck, ob das EPROM leer ist.
 */
uint8_t eprom_is_empty(void) {
    uint16_t i = 0;

    // Pruefe jedes Byte des EPROMS.
    for (i = 0; i < EPROM_SIZE; i++) {

        // Falls das gelesene Byte nicht 0xFF ist, ist das
        // EPROM nicht leer.
        if(eprom_read(i) != 0xff) {
            break;
        }

    }

    // Die obige Schleife wird verlassen, wenn eine 
    // beschriebene Speicherzelle gefunden wird oder sie das
    // ganze PROM abgesucht hat. Durch Vergleich der Ausstiegs-
    // stelle mit der Groesse erkennt man beschriebene PROMs.
    if (i < EPROM_SIZE) {
        return EPROM_NEMPTY;
    } else { // Das EPROM ist leer
        return EPROM_EMPTY;
    }

}

/*
 * Leert das EEPROM, indem in jedes Byte 0xFF geschrieben wird.
 */
void eprom_clear_eeprom(void) {

    // Falls das EPROM kein EEPROM ist, erzeuge eine
    // Fehlermeldung.
    if (eprom_get_type() != EPROM_EEPROM) {
        misc_error();
        return;
    }

    // Initialisiere den Brennvorgang.
    eprom_write_start();

    uint16_t i = 0;

    // Schreibe in jedes Byte 0xFF.
    for (i = 0; i < EPROM_SIZE; i++) {
        eprom_write(i, 0xff);
    }

    // Beende den Brennvorgang.
    eprom_write_stop();
}
