#include "misc.h"

/*
 * Erzeugt eine Fehlermeldung, indem das Programm in eine
 * Endlosschleife gefuehrt wird, in der die LED immer wieder
 * an- und ausgeschaltet wird. Eine Rueckkehr aus diesem
 * Zustand ist nur durch einen Reset des Controllers moeglich.
 */
void misc_error(void) {

    // Erzeugen einer Endlosschleife, damit die normale
    // Programmausfuehrung verhindert wird.
    while (1) {

        // Schalte die LED ein und warte 0,5 Sekunden.
        PORTD &= ~(1<<LED);
        _delay_ms(250);
        _delay_ms(250);

        // Schalte die LED aus und warte 0,5 Sekunden.
        PORTD |= (1<<LED);
        _delay_ms(250);
        _delay_ms(250);
    }

}

/*
 * Aktiviert oder deaktiviert den Output des Controllers am
 * LCD-/EPROM-Bus.
 */
void misc_set_dd_uc(uint8_t enable) {

    // Falls enable nicht 0 ist, soll der Output aktiviert
    // werden.
    if (enable) {
        DDRA = 0xff;
    } else { // Der Output soll deaktiviert werden
        DDRA = 0;
        PORTA = 0;
    }

}

/*
 * Aktiviert oder deaktiviert den Output des EPROMs am
 * LCD-/EPROM-Bus.
 */
void misc_set_dd_eprom(uint8_t enable) {

    // Falls enable nicht 0 ist, soll der Output aktiviert
    // werden.
    if (enable) {
        PORTC &= ~(1<<GOE);
    } else { // Der Output soll deaktiviert werden
        PORTC |= (1<<GOE);
    }

}

/*
 * Aktiviert oder deaktiviert den Output des LCDs am
 * LCD-/EPROM-Bus.
 */
void misc_set_dd_lcd(uint8_t enable) {

    // Falls enable nicht 0 ist, soll der Output aktiviert
    // werden.
    if (enable) {
        PORTC |= (1<<RW);
    } else { // Der Output soll deaktiviert werden
        PORTC &= ~(1<<RW);
    }

}

// Speichert aktuellen Zustand des LCD-/EPROM-Busses.
uint8_t misc_data_direction = 0xff;

/*
 * Setzt die Datenrichtungen der Pins des Controllers, des LCDs
 * und des EPROMs am LCD-/EPROM-Bus, so dass nur einer der 
 * drei Teilnehmer darauf schreiben kann.
*/
void misc_set_data_direction(uint8_t dd) {

    // Ist die neue Datenrichtung bereits gesetzt, tue nichts.
    if (misc_data_direction == dd) {
        return;
    }

    switch (dd) {

        // Niemand soll auf den Bus schreiben.
        case MISC_DD_OFF:

            // Deaktiviere alle Outputs auf dem Bus.
            misc_set_dd_uc(0);
            misc_set_dd_eprom(0);
            misc_set_dd_lcd(0);
            break;

        // Der Controller soll auf den Bus schreiben.
        case MISC_DD_UC:

            misc_set_dd_eprom(0);
            misc_set_dd_lcd(0);
            misc_set_dd_uc(1);
            break;

        // Das EPROM soll auf den Bus schreiben.
        case MISC_DD_EPROM:

            misc_set_dd_uc(0);
            misc_set_dd_lcd(0);
            misc_set_dd_eprom(1);
            break;

        // Das LCD soll auf den Bus schreiben.
        case MISC_DD_LCD:

            misc_set_dd_uc(0);
            misc_set_dd_eprom(0);
            misc_set_dd_lcd(1);
            break;

        // Der Parameter dd hat einen falschen Wert.
        default:
            misc_error();
            break;

    }

    // Speichere den neuen Zustand.
    misc_data_direction = dd;
}

/*
 * Wandelt das uebergebene halbe Byte (die 4 hoechstwertigen
 * Bits werden ignoriert) in das zugehoerige Hex-Zeichen im
 * ASCII-Format um, welches dann am LCD angezeigt werden kann.
 */
uint8_t misc_halfbyte_to_char(uint8_t hb) {

    // Ignoriere die 4 hoechstwertigen Bits,
    // hb hat also einen Wert zwischen 0 und 16.
    hb &= 0x0f;

    // Falls das halbe Byte kleiner als 10 ist, ist es eine
    // Ziffer in der Hex-Darstellung.
    if (hb < 10) {

        // Addiere den Offset 0x30, um aus der Zahl hb den
        // ASCII-Code zu erhalten, der hb darstellt.
        hb += 0x30;

    } else { // Das halbe Byte ist ein Buchstabe

        // Addiere den Offset 0x37, um den entsprechenden
        // Buchstaben A-F zu erhalten.
        hb += 0x37;
    }

    return hb;
}
