#include "pc.h"

// Speichert den aktuellen Zustand.
uint8_t pc_state = PC_STATE_STOP;

// Speichert die Groesse der zu empfangenen Daten.
uint16_t pc_write_size = 0;

// Speichert die Anzahl der bereits empfangenen Daten.
uint16_t pc_write_i = 0;

// Leert den Empfangspuffer.
void pc_flush(void) {

    // Solange noch Daten im Puffer sind, auslesen.
    while (usart_has_data()) {
        usart_read_char();
    }

}

/*
 * Startet das PC-Protokoll
 */
void pc_start(void) {

    // Falls das PC-Protokoll laeuft, erst stoppen.
    if (pc_state != PC_STATE_STOP) {
        pc_stop();
    }

    // Puffer leeren und in den Start-Zustand gehen.
    pc_state = PC_STATE_START;
    pc_flush();
}

/*
 * Prueft, ob Daten empfangen wurden und verarbeitet diese.
 */
void pc_process(void) {

    // Laeuft das PC-Protokoll nicht oder wurden noch keine
    // Daten empfangen, abbrechen.
    if (pc_state == PC_STATE_STOP || !usart_has_data()) {
        return;
    }

    // Empfangenes Zeichen lesen.
    uint8_t data = usart_read_char();

    // Verarbeitung abhaengig vom aktuellen Zustand.
    switch (pc_state) {

        case PC_STATE_START: // Start-Zustand

            // Wenn PC_CONNECT empfangen wurde, Version des
            // Protokolls senden in den Zustand Verbunden
            // uebergehen.
            if (data == PC_CONNECT) {
                pc_state = PC_STATE_CONNECTED;
                usart_write_char(PC_VERSION);
            }

            break;

        case PC_STATE_CONNECTED: // Zustand Verbunden

            // Auf empfangenes Zeichen reagieren.
            switch (data) {

                // Die Verbindung wird abgebaut, alle Variablen
                // zuruecksetzen, in den Start-Zustand gehen.
                case PC_DISCONNECT:
                    pc_state = PC_STATE_START;
                    pc_write_size = 0;
                    pc_write_i = 0;
                    usart_write_char(PC_DISCONNECT);
                    break;

                // Die maximale Groesse des Editors wird
                // angefordert, erst wird das MSB,
                // anschliessend das LSB gesendet.
                case PC_SIZE:
                    usart_write_char(EDITOR_MAXMEMSIZE>>8);
                    usart_write_char(
                            (uint8_t) EDITOR_MAXMEMSIZE);
                    break;

                // Information ueber das EPROM wird
                // angefordert, erst wird der Typ gesendet,
                // dann, ob es leer ist.
                case PC_INFO:

                    if (eprom_get_type() == EPROM_NEPROM) {
                        usart_write_char(PC_NEPROM);
                    } else {
                        usart_write_char(PC_EEPROM);
                    }

                    if (eprom_is_empty()) {
                        usart_write_char(PC_EMPTY);
                    } else {
                        usart_write_char(PC_NEMPTY);
                    }

                    break;

                // Das nicht-leere EEPROM soll geloescht
                // werden. Ist ein nicht-leeres EPROM
                // eingesteckt, so wird es geloescht und es
                // wird CLEARED gesendet, ansonsten wird
                // NCLEARED gesendet.
                case PC_CLEAR:

                    if (eprom_get_type() == EPROM_EEPROM &&
                            !eprom_is_empty()) {
                        eprom_clear_eeprom();
                        usart_write_char(PC_CLEARED);
                    } else {
                        usart_write_char(PC_NCLEARED);
                    }

                    break;

                // Der Editor soll geleert werden, also wird er
                // geleert und FLUSHED gesendet.
                case PC_FLUSH:
                    editor_clear();
                    usart_write_char(PC_FLUSHED);
                    break;

                // Der PC fordert die Daten aus dem EPROM an.
                case PC_READ:

                    // Das EPROM auslesen.
                    editor_read_eprom();

                    // Groesse der EPROM-Daten senden, 
                    // MSB zuerst.
                    usart_write_char(editor_end_pointer>>8);

                    // Stabilisierung abwarten
                    _delay_ms(60);

                    // LSB uebertragen
                    usart_write_char(editor_end_pointer);

                    uint16_t i = 0;

                    // Daten uebertragen, dabei jeweils warten.
                    for (i = 0; i < editor_end_pointer; i++) {
                        _delay_ms(60);
                        usart_write_char(editor_buffer[i]);
                    }

                    break;

                // PC will in den Editor schreiben -- Zustand
                // wechseln und Bestaetigung senden.
                case PC_WRITE:
                    pc_state = PC_STATE_WRITE_SIZE;
                    usart_write_char(PC_WRITE_SIZE);
                    break;

                // Das EPROM soll gebrannt werden, also
                // Bestaetigung senden, das EPROM brennen und
                // dann erneut Bestaetigung senden.
                case PC_BURN:
                    usart_write_char(PC_BURNING);
                    editor_burn();
                    usart_write_char(PC_BURNED);
                    break;

            }

            break;

        // Die Groesse der zu empfangenen Daten wird empfangen.
        case PC_STATE_WRITE_SIZE:

            if (pc_write_i == 0) { // MSB zuerst

                // Zaehlvariable setzen, Datum speichern und
                // Bestaetigung senden.
                pc_write_i = 1;
                pc_write_size = data<<8;
                usart_write_char(PC_WRITE_SIZE);

            } else { // Das naechste das LSB

                // Zaehlvariable zuruecksetzen, Datum speichern
                // und Bestaetigung senden sowie Zustand
                // wechseln.
                pc_write_i = 0;
                pc_state = PC_STATE_WRITE;
                pc_write_size |= data;
                usart_write_char(PC_WRITE_NEXT);

            }

            break;

        // Die Daten fuer den Editor werden empfangen.
        case PC_STATE_WRITE:

            // Datum speichern, Zaehler erhoehen.
            editor_buffer[pc_write_i++] = data;

            if (pc_write_i >= pc_write_size) {
                // Alle Daten empfangen.

                // In den Zustand Verbunden wechseln, Zaehler
                // zuruecksetzen und Bestaetigung schicken.
                pc_state = PC_STATE_CONNECTED;
                pc_write_i = 0;
                usart_write_char(PC_WRITTEN);

            } else { // Daten stehen noch aus

                // Bestaetigung senden, weitere Daten erwarten.
                usart_write_char(PC_WRITE_NEXT);

            }

            break;

        // Unbekanntes Zeichen verwerfen.
        default:
            break;

    }

}

/*
 * Stoppt das PC-Protokoll und setzt alle Variablen zurueck.
 */ 
void pc_stop(void) {

    // Nur zuruecksetzen, wenn nicht bereits gestoppt.
    if (pc_state != PC_STATE_STOP) {

        // Falls noch verbunden, trennen.
        if (pc_state != PC_STATE_START) {
            usart_write_char(PC_DISCONNECT);
        }

        // Zuruecksetzen.
        pc_state = PC_STATE_STOP;
        pc_write_size = 0;
        pc_write_i = 0;

    }

}
