#include "twi.h"

// Speichert den aktuellen Zustand.
uint8_t twi_state = TWI_STATE_STOP;

// Speichert den zuletzt uebertragenen Wert
uint8_t twi_data = 0xff;

/*
 * Initialisiert TWI. Setzt die eigene Adresse sowie die
 * Uebertragungsgeschwindigkeit und aktiviert TWI.
 */
void twi_init(void) {

    // Setze die eigene Adresse.
    TWAR = TWI_ADDRESS<<1;

    // Setze die Uebertragungsgeschwindigkeit
    TWBR = 32;
    TWSR = 2;

    // Aktiviere TWI.
    TWCR = (1<<TWEN);
}

/*
 * Setzt die Slave-Adresse und sendet diese.
 */
void twi_set_address(uint8_t address) {

    // Nur senden, wenn im richtigen Zustand.
    if (twi_state == TWI_STATE_ADDRESS_SET) {

        // In neuen Zustand gehen
        twi_state = TWI_STATE_ADDRESS;

        // Adresse setzen und ReceiverMode aktivieren.
        TWDR = (address<<1) | 0x01;

        // Uebertragen.
        TWCR = (1<<TWEN) | (1<<TWINT);
    }

}

/*
 * Startet die Uebertragung.
 */
void twi_start(void) {

    // Nur starten, wenn im richtigen Zustand.
    if (twi_state == TWI_STATE_STOP) {
        twi_state = TWI_STATE_START;

        // Start senden.
        TWCR = (1<<TWEN) | (1<<TWSTA);
    }

}

/*
 * Stoppt die Uebertragung unabhengig vom Zustand.
 */
void twi_stop(void) {
    twi_state = TWI_STATE_STOP;

    // Stop senden, Uebertragung beenden.
    TWCR = (1<<TWEN) | (1<<TWINT) | (1<<TWSTO);
}

/*
 * Gibt den zuletzt empfangenen Wert zurueck.
 */
uint8_t twi_get_data(void) {

    // Nur im richtigen Zustand Uebertragung beenden.
    if (twi_state == TWI_STATE_READ_GET) {
        twi_stop();
    }

    // Wert zurueckgeben.
    return twi_data;
}

/*
 * Prueft, ob sich bei der Uebertragung etwas getan hat und
 * reagiert darauf.
 */
uint8_t twi_process(void) {

    // Interrupt ist aufgetreten
    if (TWCR & (1<<TWINT)) {

        // Status auslesen und verarbeiten
        switch (TWSR & 0xf8) {

            // Start wurde uebertragen
            case 0x08:

                if (twi_state == TWI_STATE_START) {

                    // Aufrufer mitteilen, dass Adresse gesetzt
                    // werden kann und Zustand wechseln.
                    twi_state = TWI_STATE_ADDRESS_SET;
                    return TWI_SET_ADDRESS;

                } else if (twi_state == TWI_STATE_ADDRESS_SET) {
                    
                    // Derselbe Interrupt, es wurde nur
                    // noch nicht die Adresse gesetzt.
                    // Aufrufer (noch einmal) mitteilen, 
                    // dass Adresse gesetzt werden kann.
                    return TWI_SET_ADDRESS;

                } else { // Im falschen Zustand

                    // Uebertragung beenden.
                    twi_stop();
                    return TWI_ERROR;
                }

                break;

            // Bestaetigung vom Slave eingetroffen, dass er
            // einen Wert senden soll.
            case 0x40:

                if (twi_state == TWI_STATE_ADDRESS) {

                    // Wert empfangen, dabei kein ACK senden.
                    twi_state = TWI_STATE_READ;
                    TWCR = (1<<TWEN) | (1<<TWINT);

                } else { // Im falschen Zustand

                    // Uebertragung beenden.
                    twi_stop();
                    return TWI_ERROR;
                }

                break;

            // Wert vom Slave empfangen, kein ACK gesendet.
            case 0x58:

                if (twi_state == TWI_STATE_READ) {
                    twi_state = TWI_STATE_READ_GET;
                    twi_data = TWDR;
                    return TWI_GET_DATA;
                } else if (twi_state == TWI_STATE_READ_GET) {
                    
                    // Derselbe Interrupt, Wert noch nicht
                    // gelesen. Wiederholt senden, dass ein 
                    // Wert gelesen werden kann.
                    return TWI_GET_DATA;

                } else { // Im falschen Zustand

                    // Uebertragung beenden.
                    twi_stop();
                    return TWI_ERROR;
                }

                break;

            // Nicht erwarteter Status, Fehler annehmen.
            default:

                // Uebertragung beenden.
                twi_stop();
                return TWI_ERROR;
                break;

        }

    }

    // Mitteilen, dass nichts passiert ist.
    return TWI_NOP;
}

/*
 * Initiert eine Uebertragung und gibt eine Statusmeldung
 * zurueck.
 */
uint8_t twi_read(uint8_t address) {
    uint16_t i = 0;

    // Uebertragung beginnen.
    twi_start();

    // Solange kein Timeout vorliegt, weiter arbeiten.
    while (i < TWI_TIMEOUT) {

        // 
        switch (twi_process()) {

            // Adresse kann gesetzt werden.
            case TWI_SET_ADDRESS:

                // Timeout-Zaehler zuruecksetzen.
                i = 0;

                // Adresse setzen und uebertragen.
                twi_set_address(address);
                break;

            // Wert wurde empfangen, kann abgefragt werden.
            case TWI_GET_DATA:

                // Dem Aufrufer mitteilen, dass ein Wert
                // angekommen ist.
                return TWI_GET_DATA;
                break;

            // Noch nichts passiert
            case TWI_NOP:

                // Timeout-Zaehler erhoehen.
                i++;
                break;

            // Fehler oder falscher Wert wurde zurueckgegeben
            default:

                // Uebertragung stoppen, Fehler zurueckgeben.
                twi_stop();
                return TWI_ERROR;
                break;

        }

    }

    // Timeout wurde erreicht, Uebertragung beenden, Fehler
    // zurueckgeben.
    twi_stop();
    return TWI_ERROR;
}
