#include "usi.h"

// Speichert den aktuellen Zustand.
uint8_t usi_state = USI_STATE_START;

// Speichert den zuletzt uebertragenen Wert
uint8_t usi_data;

/*
    Initialisiert USI. Aktiviert USI und konfiguriert
    SDA und SCL.
*/
void usi_init(void) {
    USICR = (1<<USIWM1) | (1<<USIWM0) | (1<<USICS1);
    PORTB |= (1<<SDA) | (1<<SCL);
    DDRB &= ~(1<<SDA);
    DDRB |= (1<<SCL);
}

// Gibt zurueck, ob Startcondition empfangen wurde.
uint8_t usi_is_start(void) {
    return (USISR & (1<<USISIF));
}

/*
    Gibt zurueck, ob Overflow passiert ist. Es wurden also 8 Bit
    empfangen.
*/
uint8_t usi_is_overflow(void) {
    return (USISR & (1<<USIOIF));
}

// Gibt zurueck, ob Stopcondition empfangen wurde.
uint8_t usi_is_stop(void) {
    return (USISR & (1<<USIPF));
}

// Gibt zurueck, ob eine Kollision bemerkt wurde.
uint8_t usi_is_collision(void) {
    return (USISR & (1<<USIDC));
}

// Setzt den zu uebertragenen Wert und sendet diesen.
void usi_set_data(uint8_t data) {

    // Nur senden, wenn im richtigen Zustand.
    if (usi_state == USI_STATE_READ_SET) {
        usi_state = USI_STATE_READ;
        USIDR = data;
        USISR |= (1<<USIOIF);
    }

}

// Gibt den zuletzt empfangenen Wert zurueck.
uint8_t usi_get_data(void) {

    // Nur bestaetigen, wenn im richtigen Zustand.
    if (usi_state == USI_STATE_WRITE_GET) {
        usi_state = USI_STATE_WRITE_ACK;
        USISR |= (1<<USIOIF);
    }

    return usi_data;
}

/*
    Prueft, ob sich bei der Uebertragung etwas getan hat und
    reagiert darauf.
*/
uint8_t usi_process(void) {
    /*
        Warten, um doppelte Abarbeitung desselben Events zu
        verhindern
    */
    _delay_ms(8);

    if (usi_is_start()) { // Startcondition
        usi_state = USI_STATE_ADDRESS;
        USISR = (1<<USISIF);
    } else if (usi_is_overflow()) { // Overflow

        // Abhaengig vom Zustand reagieren
        switch (usi_state) {

            // Adresse wurde uebertragen
            case USI_STATE_ADDRESS:

                if ((USIDR>>1) == USI_ADDRESS) {
                        // Eigene Adresse wurde angesprochen

                    if (USIDR & 0x01) {
                            // Wert soll geschrieben werden
                        usi_state = USI_STATE_READ_ACK;
                    } else {
                            // Wert soll gelesen werden
                        usi_state = USI_STATE_WRITE_ACK;
                    }

                    // ACK senden
                    DDRB |= (1<<SDA);
                    USIDR = 0x00;
                    USISR = 0x0e | (1<<USIOIF);
                } else { // Im falschen Zustand
                    usi_reset();
                }

                break;

            // ACK wurde gesendet, Wert soll nun gesetzt werden
            case USI_STATE_READ_ACK:

                if (USIDR & 0x01) { // Kein ACK
                    usi_reset();
                } else { // ACK

                    // Wert muss nun gesetzt werden
                    usi_state = USI_STATE_READ_SET;
                    DDRB |= (1<<SDA);
                    return USI_SET_DATA;
                }

                break;

            // Wert muss immer noch gesetzt werden
            case USI_STATE_READ_SET:
                return USI_SET_DATA;
                break;

            // ACK kann empfangen werden
            case USI_STATE_READ:
                usi_state = USI_STATE_READ_ACK;
                DDRB &= ~(1<<SDA);
                USISR = 0x0e;
                USISR = (1<<USIOIF);
                break;

            // ACK wurde gesendet, Wert kann empfangen werden
            case USI_STATE_WRITE_ACK:
                usi_state = USI_STATE_WRITE;
                DDRB &= ~(1<<SDA);
                USISR = (1<<USIOIF);
                break;

            /*
                Wert wurde empfangen, ACK vorbereiten,
                Wert muss erst abgeholt werden
            */
            case USI_STATE_WRITE:
                usi_state = USI_STATE_WRITE_GET;
                usi_data = USIDR;
                DDRB |= (1<<SDA);
                USIDR = 0x01;
                USISR = 0x0e;
                return USI_GET_DATA;
                break;

            // Wert muss immer noch abgeholt werden
            case USI_STATE_WRITE_GET:
                return USI_GET_DATA;
                break;

            default: // Unbehandelter Fall, Fehler annehmen
                usi_reset();
                break;

        }

    } else if (usi_is_stop()) { // Stopcondition
        // Zuruecksetzen
        usi_reset();
    } else if (usi_is_collision()) { // Kollision
        // Zuruecksetzen
        usi_reset();
    }

    // Nichts interessantes passiert
    return USI_NOP;
}

// USI zuruecksetzen.
void usi_reset(void) {
    usi_state = USI_STATE_START;
    DDRB &= ~(1<<SDA);
    USISR = (1<<USISIF) | (1<<USIOIF) | (1<<USIPF) | (1<<USIDC);
}
